@@ -20,7 +20,13 @@
 
 GucVariable conf[MAX_SETTINGS];
 static void DefineIntGUC(GucVariable *conf);
+static void DefineIntGUCWithCheck(GucVariable *conf, GucIntCheckHook check);
 static void DefineBoolGUC(GucVariable *conf);
+static void DefineEnumGUC(GucVariable *conf, const struct config_enum_entry *options);
+
+/* Check hooks to ensure histogram_min < histogram_max */
+static bool check_histogram_min(int *newval, void **extra, GucSource source);
+static bool check_histogram_max(int *newval, void **extra, GucSource source);
 
 /*
  * Define (or redefine) custom GUC variables.
@@ -29,6 +35,7 @@ void
 init_guc(void)
 {
 	int i = 0;
+
 	conf[i] = (GucVariable) {
 		.guc_name = "pg_stat_monitor.pgsm_max",
 		.guc_desc = "Sets the maximum size of shared memory in (MB) used for statement's metadata tracked by pg_stat_monitor.",
@@ -44,7 +51,7 @@ init_guc(void)
 	conf[i] = (GucVariable) {
 		.guc_name = "pg_stat_monitor.pgsm_query_max_len",
 		.guc_desc = "Sets the maximum length of query.",
-		.guc_default = 1024,
+		.guc_default = 2048,
 		.guc_min = 1024,
 		.guc_max = INT_MAX,
 		.guc_unit = 0,
@@ -53,18 +60,6 @@ init_guc(void)
 	};
 	DefineIntGUC(&conf[i++]);
 
-	conf[i] = (GucVariable) {
-		.guc_name = "pg_stat_monitor.pgsm_enable",
-		.guc_desc = "Enable/Disable statistics collector.",
-		.guc_default = 1,
-		.guc_min = 0,
-		.guc_max = 0,
-		.guc_restart = false,
-		.guc_unit = 0,
-		.guc_value = &PGSM_ENABLED
-	};
-	DefineBoolGUC(&conf[i++]);
-
 	conf[i] = (GucVariable) {
 		.guc_name = "pg_stat_monitor.pgsm_track_utility",
 		.guc_desc = "Selects whether utility commands are tracked.",
@@ -123,7 +118,7 @@ init_guc(void)
 		.guc_unit = 0,
 		.guc_value = &PGSM_HISTOGRAM_MIN
 	};
-	DefineIntGUC(&conf[i++]);
+	DefineIntGUCWithCheck(&conf[i++], check_histogram_min);
 
 	conf[i] = (GucVariable) {
 		.guc_name = "pg_stat_monitor.pgsm_histogram_max",
@@ -135,14 +130,14 @@ init_guc(void)
 		.guc_unit = 0,
 		.guc_value = &PGSM_HISTOGRAM_MAX
 	};
-	DefineIntGUC(&conf[i++]);
+	DefineIntGUCWithCheck(&conf[i++], check_histogram_max);
 
 	conf[i] = (GucVariable) {
 		.guc_name = "pg_stat_monitor.pgsm_histogram_buckets",
 		.guc_desc = "Sets the maximum number of histogram buckets",
 		.guc_default = 10,
 		.guc_min = 2,
-		.guc_max = INT_MAX,
+		.guc_max = MAX_RESPONSE_BUCKET,
 		.guc_restart = true,
 		.guc_unit = 0,
 		.guc_value = &PGSM_HISTOGRAM_BUCKETS
@@ -185,12 +180,39 @@ init_guc(void)
 	};
 	DefineBoolGUC(&conf[i++]);
 
+	conf[i] = (GucVariable) {
+		.guc_name = "pg_stat_monitor.pgsm_track",
+		.guc_desc = "Selects which statements are tracked by pg_stat_monitor.",
+		.n_options = 3,
+		.guc_default = PGSM_TRACK_TOP,
+		.guc_min = PSGM_TRACK_NONE,
+		.guc_max = PGSM_TRACK_ALL,
+		.guc_restart = false,
+		.guc_unit = 0,
+		.guc_value = &PGSM_TRACK
+	};
+	for (int j = 0; j < conf[i].n_options; ++j) {
+		strlcpy(conf[i].guc_options[j], track_options[j].name, sizeof(conf[i].guc_options[j]));
+	}
+	DefineEnumGUC(&conf[i++], track_options);
+
+	conf[i] = (GucVariable) {
+		.guc_name = "pg_stat_monitor.pgsm_extract_comments",
+		.guc_desc = "Enable/Disable extracting comments from queries.",
+		.guc_default = 0,
+		.guc_min = 0,
+		.guc_max = 0,
+		.guc_restart = false,
+		.guc_unit = 0,
+		.guc_value = &PGSM_EXTRACT_COMMENTS
+	};
+	DefineBoolGUC(&conf[i++]);
 
 #if PG_VERSION_NUM >= 130000
 	conf[i] = (GucVariable) {
 		.guc_name = "pg_stat_monitor.pgsm_track_planning",
 		.guc_desc = "Selects whether planning statistics are tracked.",
-		.guc_default = 1,
+		.guc_default = 0,
 		.guc_min = 0,
 		.guc_max = 0,
 		.guc_restart = false,
@@ -201,9 +223,9 @@ init_guc(void)
 #endif
 }
 
-static void
-DefineIntGUC(GucVariable *conf)
+static void DefineIntGUCWithCheck(GucVariable *conf, GucIntCheckHook check)
 {
+	conf->type = PGC_INT;
 	DefineCustomIntVariable(conf->guc_name,
 							conf->guc_desc,
 							NULL,
@@ -213,13 +235,21 @@ DefineIntGUC(GucVariable *conf)
 							conf->guc_max,
 							conf->guc_restart ? PGC_POSTMASTER : PGC_USERSET,
 							conf->guc_unit,
-							NULL,
+							check,
 							NULL,
 							NULL);
 }
+
+static void
+DefineIntGUC(GucVariable *conf)
+{
+	DefineIntGUCWithCheck(conf, NULL);
+}
+
 static void
 DefineBoolGUC(GucVariable *conf)
 {
+	conf->type = PGC_BOOL;
 	DefineCustomBoolVariable(conf->guc_name,
 							conf->guc_desc,
 							NULL,
@@ -232,9 +262,39 @@ DefineBoolGUC(GucVariable *conf)
 							 NULL);
 }
 
+static void
+DefineEnumGUC(GucVariable *conf, const struct config_enum_entry *options)
+{
+	conf->type = PGC_ENUM;
+	DefineCustomEnumVariable(conf->guc_name,
+                             conf->guc_desc,
+                             NULL,
+                             conf->guc_value,
+                             conf->guc_default,
+                             options,
+                             conf->guc_restart ? PGC_POSTMASTER : PGC_USERSET,
+                             0,
+                             NULL,
+                             NULL,
+                             NULL);
+}
+
 GucVariable*
 get_conf(int i)
 {
-	return  &conf[i];
+	return &conf[i];
 }
 
+static bool check_histogram_min(int *newval, void **extra, GucSource source)
+{
+	/*
+	 * During module initialization PGSM_HISTOGRAM_MIN is initialized before
+	 * PGSM_HISTOGRAM_MAX, in this case PGSM_HISTOGRAM_MAX will be zero.
+	 */
+	return (PGSM_HISTOGRAM_MAX == 0 || *newval < PGSM_HISTOGRAM_MAX);
+}
+
+static bool check_histogram_max(int *newval, void **extra, GucSource source)
+{
+	return (*newval > PGSM_HISTOGRAM_MIN);
+}
